# Copyright © 2015-2018 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging
import pathlib
import tomllib

from configparser import ConfigParser
from datetime import datetime
from os import access, chmod, environ, listdir, makedirs, walk, X_OK
from os.path import abspath, exists, isdir, join, dirname
from shutil import copy

from pypi2deb import VERSION, OVERRIDES_PATH, PROFILES_PATH, TEMPLATES_PATH
from pypi2deb.tools import execute

from jinja2 import Environment, FileSystemLoader
from debian.changelog import Changelog, Version, get_maintainer
try:
    from simplejson import load, dump
except ImportError:
    from json import load, dump

from dhpython import PKG_PREFIX_MAP
from dhpython.pydist import guess_dependency, parse_pydep


log = logging.getLogger('pypi2deb')
# The location of sphinx-based documentation can be configured via the
# environment variable PYPI2DEB_SPHINX_DIRS (colon-separated directories).
# Disable sphinx support by assigning an empty string to this environment
# variable.
try:
    SPHINX_DIR = set(environ['PYPI2DEB_SPHINX_DIRS'].split(":"))
except KeyError:
    SPHINX_DIR = {'docs', 'doc', 'doc/build'}
INTERPRETER_MAP = {value: key for key, value in PKG_PREFIX_MAP.items()}
VERSIONED_I_MAP = {'python': 'python2'}
DESC_STOP_KEYWORDS = {'changelog', 'changes', 'license', 'requirements',
                      'installation'}


def _copy_static_files(src_dir, debian_dir):
    """Copy static templates/overrides from source to destination"""
    for root, dirs, file_names in walk(src_dir):
        for fn in file_names:
            if fn.endswith(('.tpl', '.swp')):
                continue
            dst_dir = root.replace(src_dir, debian_dir)
            if not exists(dst_dir):
                makedirs(dst_dir)
            if not exists(join(dst_dir, fn)):
                copy(join(root, fn), join(dst_dir, fn))


async def debianize(dpath, ctx, profile=None):
    update_ctx(dpath, ctx)

    setupcfg_fpath = join(dpath, 'setup.cfg')
    if exists(setupcfg_fpath):
        upstream_cfg = ConfigParser()
        upstream_cfg.read(setupcfg_fpath)
        if 'py2dsp' in upstream_cfg:
            ctx.update(upstream_cfg['py2dsp'].items())

    override_paths = [TEMPLATES_PATH]

    # profile overrides
    if profile:
        if isdir(profile):  # --profile a_dir
            override_paths.append(profile)
        elif exists(profile):  # --profile a_file
            with open(profile) as fp:
                set_keys = [key for (key, value) in ctx.items() if isinstance(value, set)]
                ctx.update(load(fp))
                # recover sets from Json lists
                for set_key in set_keys:
                    if not isinstance(ctx[set_key], set):
                        ctx[set_key] = set(ctx[set_key])
        else:  # --profile name
            profile_dpath = join(PROFILES_PATH, profile)
            if isdir(profile_dpath):
                override_paths.append(profile_dpath)

    # package specific overrides... in global overrides dir
    o_dpath = join(OVERRIDES_PATH, ctx['name'].lower())
    isdir(o_dpath) and override_paths.append(o_dpath)
    # ... in local ./overrides
    o_dpath = join('overrides', ctx['name'].lower())
    isdir(o_dpath) and override_paths.append(o_dpath)

    # handle overrides
    debian_dir = join(dpath, 'debian')
    for o_dpath in override_paths:
        fpath = join(o_dpath, 'ctx.json')
        if exists(fpath):
            with open(fpath) as fp:
                set_keys = [key for (key, value) in ctx.items() if isinstance(value, set)]
                ctx.update(load(fp))
                # recover sets from Json lists
                for set_key in set_keys:
                    if not isinstance(ctx[set_key], set):
                        ctx[set_key] = set(ctx[set_key])
        # invoke pre hooks
        fpath = abspath(join(o_dpath, 'hooks', 'pre'))
        if exists(fpath) and access(fpath, X_OK):
            _dump_ctx(ctx)
            code = await execute([fpath, ctx['src_name'],
                                       ctx['version'], ctx['debian_revision']],
                                      cwd=dpath)
            if code != 0:
                raise Exception("pre hook for %s failed with %d return code" % (
                                ctx['name'], code))
    for o_dpath in reversed(override_paths):
        # copy static files
        deb_dpath = join(o_dpath, 'debian')
        if isdir(deb_dpath):
            _copy_static_files(deb_dpath, debian_dir)

    for key in ('vcs_src', 'vcs_browser', 'uploaders'):
        if key in ctx:
            ctx[key] = ctx[key].format(**ctx)
    ctx['debian_version'] = "{}-{}".format(ctx['version'], ctx['debian_revision'])

    # Jinja setup: set templates directories
    templates_dir = [dpath]  # use existing dir as a template dir as well
    templates_dir.extend(override_paths)
    templates_dir.append(TEMPLATES_PATH)
    env = Environment(loader=FileSystemLoader(templates_dir))

    # render debian dir files (note that order matters)
    docs(dpath, ctx, env)
    control(dpath, ctx, env)
    rules(dpath, ctx, env)
    chmod(join(dpath, 'debian', 'rules'), 0o755)
    initial_release = await changelog(dpath, ctx, env)
    if initial_release:
        itp_mail(dpath, ctx, env)
    copyright(dpath, ctx, env)
    watch(dpath, ctx, env)
    # Currently only Github is supported for DEP-12
    if 'github' in ctx:
        upstream__metadata(dpath, ctx, env)
    clean(dpath, ctx, env)

    # invoke post hooks
    for o_dpath in override_paths:
        fpath = join(o_dpath, 'hooks', 'post')
        if exists(fpath) and access(fpath, X_OK):
            _dump_ctx(ctx)
            code = await execute([fpath, ctx['src_name'],
                                       ctx['version'], ctx['debian_revision'],
                                       VERSION],
                                      cwd=dpath)
            if code != 0:
                raise Exception("post hook for %s failed with %d return code" % (
                                ctx['name'], code))


def update_ctx(dpath, ctx):
    ctx.setdefault('exports', {})
    ctx.setdefault('build_depends', set())
    maintainer, email = get_maintainer()
    ctx['creator'] = '{} <{}>'.format(maintainer, email)
    if 'maintainer' not in ctx:
        ctx['maintainer'] = '{} <{}>'.format(maintainer, email)
    if 'debian_revision' not in ctx:
        ctx['debian_revision'] = '0~pypi2deb'

    ctx['binary_arch'] = 'all'
    ctx.setdefault('clean_files', set())
    for root, dirs, file_names in walk(dpath):
        if any(fname.endswith(('.c', '.cpp', '.pyx')) for fname in file_names):
            ctx['binary_arch'] = 'any'

            for fname in file_names:
                if fname.endswith('.pyx'):
                    if 'python3' in ctx['interpreters']:
                        ctx['build_depends'].add('cython3')
                    for ext in ('c', 'cpp'):
                        fname_c = fname[:-3] + ext
                        if fname_c in file_names:
                            ctx['clean_files'].add(join(root.replace(dpath, '.'), fname_c))


def _dump_ctx(ctx):
    """dump ctx in JSON format so that hooks can use it"""
    try:
        fpath = join(ctx['root'], '{src_name}_{version}-{debian_revision}.ctx'.format(**ctx))
        serializable_ctx = {key: (value if not isinstance(value, set) else tuple(value))
                            for (key, value) in ctx.items()}
        with open(fpath, 'w') as fp:
            dump(serializable_ctx, fp, indent=' ')
    except Exception:
        log.debug('cannot dump ctx', exc_info=True)


def docs(dpath, ctx, env):
    docs = ctx.setdefault('docs', {})
    for path in SPHINX_DIR:
        if exists(join(dpath, path, 'Makefile')) and exists(join(dpath, path, 'conf.py')):
            docs['sphinx_dir'] = path
            ctx['build_depends'].add('python3-sphinx')
            docs.setdefault('files', []).append('.pybuild/docs/*')
    for fn in listdir(dpath):
        if fn.lower().startswith('readme'):
            docs.setdefault('files', []).append(fn)
        if fn.lower() == 'examples':
            docs['examples_dir'] = fn
    if docs:
        if 'sphinx_dir' in docs:
            # i.e. we have binary packages with docs
            docs_pkg = 'python-{}-doc'
        elif 'python3' in ctx['interpreters']:
            docs_pkg = 'python3-{}'
        docs_pkg = docs_pkg.format(ctx['src_name'])
        if 'examples_dir' in docs:
            # TODO: should we extend this file only if it exists?
            with open(join(dpath, 'debian', docs_pkg + '.examples'), 'w') as fp:
                fp.write("{}/*\n".format(docs['examples_dir']))
        if 'files' in docs:
            # TODO: should we extend this file only if it exists?
            with open(join(dpath, 'debian', docs_pkg + '.docs'), 'w') as fp:
                for fn in docs['files']:
                    fp.write("{}\n".format(fn))


def _render_template(func):
    name = func.__name__.replace('__', '/')

    def _template(dpath, ctx, env, *args, **kwargs):
        fpath = join(dpath, 'debian', name)
        if exists(fpath):
            log.debug('debian/%s already exist, skipping', name)
            return
        ctx = func(dpath, ctx, env, *args, **kwargs)
        tpl = env.get_template('debian/{}.tpl'.format(name))

        if not isdir(dirname(fpath)):
            makedirs(dirname(fpath))

        with open(fpath, 'w', encoding='utf-8') as fp:
            fp.write(tpl.render(ctx))
    return _template


@_render_template
def control(dpath, ctx, env):
    desc = []
    code_line = False
    first_line = True
    for line in ctx['description'].split('\n'):
        if first_line:
            if line.lower() == ctx['name'].lower():
                continue
            if not line.strip().replace('=', '').replace('-', '').replace('~', ''):
                continue
            first_line = False
        if not line.strip():
            desc.append(' .')
        else:
            if line.startswith(('* ', '>>> ', '... ', '.. ', '$ ')) \
               or code_line or line == '...':
                if line.startswith('>>> '):
                    # next line should get extra space char as well
                    code_line = True
                elif code_line:
                    code_line = False
                line = ' ' + line
            elif line.strip().lower() in DESC_STOP_KEYWORDS:
                break
            line = line.replace('\t', '    ')
            desc.append(' ' + line)
    for key, value in {
        'short_desc': ctx['summary'][:80],
        'long_desc': '\n'.join(desc),
    }.items():
        if key not in ctx:
            ctx[key] = value

    req = set()
    if 'requires' in ctx:
            for impl in ctx['interpreters']:
                impl = INTERPRETER_MAP.get(impl, impl)
                try:
                    dependency = guess_dependency(impl, line)
                    if dependency:
                        ctx['build_depends'].add(dependency)
                except Exception as err:
                    log.warn('cannot parse build dependency: %s', err)
    else:
        for i in listdir(dpath):
            if i.endswith('.egg-info') and exists(join(dpath, i, 'requires.txt')):
                req.add(join(dpath, i, 'requires.txt'))
            if i == 'requirements.txt':
                req.add(join(dpath, 'requirements.txt'))

        for fpath in req:
            for impl in ctx['interpreters']:
                impl = INTERPRETER_MAP.get(impl, impl)
                try:
                    for i in parse_pydep(impl, fpath)['depends']:
                        ctx['build_depends'].add(i)
                except Exception as err:
                    log.warn('cannot parse build dependency: %s', err)

    pyproject_file = pathlib.Path(dpath, 'pyproject.toml')
    if pyproject_file.exists():
        pyproject_toml = tomllib.loads(pyproject_file.read_text())
        if 'build-system' in pyproject_toml:
            # https://pip.pypa.io/en/stable/reference/build-system/pyproject-toml/#fallback-behaviour
            if 'build-backend' not in pyproject_toml['build-system']:
                log.info("Unable to detect a build backend in pyproject.toml, falling back to setuptools")
                pyproject_toml['build-system']['build-backend'] = \
                    "setuptools.build_meta:__legacy__"
            ctx['pybuild_depends'] = 'pybuild-plugin-pyproject'
            # static list of build backends and what package to install to use them
            backends = {
                'flit': 'flit',
                'hatchling': 'python3-hatchling',
                'mesonpy': 'python3-mesonpy',
                'pdm': 'python3-pdm-pep517',
                'poetry': 'python3-poetry-core',
                'setuptools': 'python3-setuptools',
                'sipbuild': 'python3-sipbuild',
            }
            for _backend, _package in backends.items():
                if _backend in pyproject_toml['build-system']['build-backend']:
                    ctx['pybuild_depends'] += f', {_package}'
        else:
            log.info("Unable to detect a build system via pyproject.toml, falling back to setup.py")

    # either there's only a setup.py or we couldnt detect a build backend via pyproject.toml
    if exists(join(dpath, 'setup.py')) and 'pybuild_depends' not in ctx:
        ctx['pybuild_depends'] = 'dh-python'
        with open(join(dpath, 'setup.py')) as fp:
            for line in fp:
                if line.startswith('#'):
                    continue
                if 'setuptools' in line:
                    for interpreter in ctx['interpreters']:
                        ctx['build_depends'].add('{}-setuptools'.format(interpreter))

    if 'python3' in ctx['interpreters']:
        ctx['build_depends'].add(
            'python3-all%s' % ('-dev' if ctx['binary_arch'] == 'any' else ''))
    if 'pypy' in ctx['interpreters']:
        ctx['build_depends'].add('pypy')

    return ctx


@_render_template
def rules(dpath, ctx, env):
    ctx['with'] = ','.join(VERSIONED_I_MAP.get(i, i) for i in ctx['interpreters'])
    if ctx.get('docs', {}).get('sphinx_dir'):
        ctx['with'] += ',sphinxdoc'

    # if package install a script in /usr/bin/ - ship it only in python3-foo package
    if exists(join(dpath, 'setup.py')) and len(ctx['interpreters']) > 1:
        with open(join(dpath, 'setup.py')) as fp:
            for line in fp:
                if 'console_scripts' in line:
                    for interpreter in ctx['interpreters']:
                        if interpreter == 'python3':
                            continue
                        ipreter = VERSIONED_I_MAP.get(interpreter, interpreter)
                        ctx['exports']['PYBUILD_AFTER_INSTALL_{}'.format(ipreter)] = 'rm -rf {destdir}/usr/bin/'
                    break

    return ctx


async def changelog(dpath, ctx, env):
    change = ctx.get('message', f"Initial release, autogenerated by py2dsp/{VERSION}; Closes: #NNNNNN")
    version = ctx['debian_version']
    distribution = ctx.get('distribution', 'UNRELEASED')

    fpath = join(dpath, 'debian', 'changelog')
    if exists(fpath):
        with open(fpath, encoding='utf-8') as fp:
            line = fp.readline()
            if ctx['version'] in line or 'UNRELEASED' in line:
                log.debug('changelog doesn\'t need an update')
                return
            else:
                await execute(['dch', '--force-distribution', '--distribution', distribution,
                                    '--newversion', version, '-m', change], cwd=dpath)
        return

    now = datetime.utcnow()
    changelog = Changelog()
    changelog.new_block(package=ctx['src_name'],
                        version=Version(version),
                        distributions=distribution,
                        urgency='low',
                        author=ctx['creator'],
                        date=now.strftime('%a, %d %b %Y %H:%M:%S +0000'))
    changelog.add_change('')
    changelog.add_change('  * {}'.format(change))
    changelog.add_change('')

    with open(fpath, 'w', encoding='utf-8') as fp:
        changelog.write_to_open_file(fp)
    return True


@_render_template
def copyright(dpath, ctx, env):
    if not ctx.get('deb_copyright'):
        ctx['deb_copyright'] = "{} {}".format(datetime.now().year, ctx['creator'])
    if ctx['license_name'] in ('Apache 2', 'Apache 2.0'):
        ctx['license_name'] = 'Apache2'
        ctx['license'] = ' See /usr/share/common-licenses/Apache-2.0'
    if not ctx.get('deb_license_name'):
        ctx['deb_license_name'] = ctx['license_name']
        ctx['deb_license'] = ctx.get('license', '')
    if not ctx.get('deb_license') and ctx['deb_license_name']:
        ctx['deb_license'] = ''
        fpath = '/usr/share/common-licenses/{}'.format(ctx['deb_license_name'])
        if exists(fpath):
            license = []
            with open(fpath, encoding='utf-8') as fp:
                for line in fp:
                    line = line.rstrip()
                    license.append(' {}'.format(line) if line else ' .')
            ctx['deb_license'] = '\n'.join(license)

    ctx.setdefault('copyright', '')

    if not ctx.get('license'):
        license = []
        for fn in listdir(dpath):
            if not fn.lower().startswith('license'):
                continue
            with open(join(dpath, fn), 'r') as fp:
                for line in fp:
                    line = line.rstrip()
                    if not line:
                        license.append(' .')
                    else:
                        license.append(' ' + line)
                        if not ctx['copyright'] and line.lower().startswith('copyright '):
                            ctx['copyright'] = line[10:]
            if license:
                ctx['license'] = '\n'.join(license)
            break
    if not ctx.get('copyright'):
        ctx['copyright'] = ctx['author']
    return ctx


@_render_template
def watch(dpath, ctx, env):
    return ctx

@_render_template
def upstream__metadata(dpath, ctx, env):
    """Render debian/upstream/metadata according to DEP-12."""
    return ctx


def itp_mail(dpath, ctx, env):
    fpath = join(ctx['root'], '{src_name}_{version}-{debian_revision}.mail'.format(**ctx))
    if exists(fpath):
        return
    tpl = env.get_template('itp.mail')
    with open(fpath, 'w', encoding='utf-8') as fp:
        fp.write(tpl.render(ctx))


def clean(dpath, ctx, env):
    old = set()
    fpath = join(dpath, 'debian', 'clean')
    if exists(fpath):
        with open(fpath) as fp:
            for line in fp:
                old.add(line.strip())

    new = set()
    for i in ctx['clean_files']:
        if i not in old:
            new.add(i)

    if new:
        with open(fpath, 'a') as fp:
            for fn in new:
                fp.write("\n{}".format(fn))



# Copyright © 2015-2018 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging
from functools import wraps
from pypi2deb.cache import load as _cache_load, dump as _cache_dump

log = logging.getLogger('pypi2deb')


def cache(ttl=3600, key=None, prefix=None):
    """Cache decorated function's result.

    :param key: static cache key
    :param prefix: prepend it to the autogenerated key
    """

    def _cache(func):
        func_name = func.__name__

        @wraps(func)
        def __cache(*args, **kwargs):
            if key:
                cache_key = key
            else:
                cache_key = "%s:%s:%s" % (func_name, args, kwargs)
            if prefix:
                cache_key = "%s:%s" % (prefix, cache_key)

            res = _cache_load(cache_key)
            if res is None:
                res = func(*args, **kwargs)
                if res is not None:
                    _cache_dump(cache_key, res, ttl)

            return res
        return __cache
    return _cache



# Copyright © 2015-2018 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import asyncio
import logging
import os
import re
import tarfile
from datetime import datetime
from os.path import exists, join
from shlex import split
from shutil import rmtree
from pypi2deb.decorators import cache
from dhpython.pydist import load, normalize_name


FILENAME_RE = re.compile(r'''
    (?:.*/)?
    (?P<name>[a-zA-Z-].*)
    [-_]
    (?P<version>[0-9][A-Za-z0-9\-.+]*?)
    (?:
        \.
        (?:orig\.)?
        (?P<extension>(?:tar(?:\.[a-z0-9]+)?)|(?:zip))
    )?$
''', re.VERBOSE)
log = logging.getLogger('pypi2deb')


def unpack(fpath, destdir='.', dname=None):
    if dname:
        dst_dpath = join(destdir, dname)
        if exists(dst_dpath):
            log.debug('{} already exists, no need to unpack'.format(dname))
            return dst_dpath
    else:
        dst_dpath = None

    fname = fpath.rsplit('/', 1)[-1]
    if '.tar.' in fname:
        with tarfile.open(fpath, 'r') as tar:
            members = []
            for memb in tar.getmembers():
                if memb.name.startswith('/') or '../' in memb.name:
                    log.warn('skipping invalid file name: %s', memb.name)
                    continue
                members.append(memb)

            dirname = members[0].name
            if not members[0].isdir():
                destdir = dst_dpath = join(destdir, dname or 'extracted')
                dirname = dname = None
                if exists(destdir):
                    log.debug('{} already exists, no need to unpack'.format(destdir))
                    return destdir
            elif not dst_dpath:
                dst_dpath = join(destdir, dirname)

            tar.extractall(destdir, members)
        if dname and dname != dirname:
            os.rename(join(destdir, dirname), dst_dpath)

        upstream_debian = os.path.join(dst_dpath, 'debian')
        if os.path.exists(upstream_debian):
            rmtree(upstream_debian)
        return dst_dpath


def parse_filename(name):
    match = FILENAME_RE.match(name)
    return match.groupdict() if match else {}


async def execute(command, cwd=None, env=None, log_output=None):
    """Execute external shell commad.

    :param cdw: currennt working directory
    :param env: environment
    :param log_output:
        * opened log file or path to this file, or
        * None if output should be redirectored to stdout/stderr
    """
    env = env or os.environ
    close = False
    if log_output:
        if isinstance(log_output, str):
            close = True
            log_output = open(log_output, 'a', encoding='utf-8')
        log_output.write('\n# command executed on {}'.format(datetime.now().isoformat()))
        log_output.write('\n$ {}\n'.format(command))
        log_output.flush()

    if isinstance(command, str):
        command = split(command)
    log.debug('invoking: %s in %s', command, cwd)

    create = asyncio.create_subprocess_exec(*command, stdout=log_output, stderr=log_output,
                                            cwd=cwd, env=env)
    proc = await create
    await proc.wait()
    close and log_output.close()

    return proc.returncode


def pkg_name(name):
    names = _load_package_names()
    name = normalize_name(name).lower()
    if name in names:
        return names[name]
    result = name.lower().replace('-python', '').replace('python-', '')
    if result.endswith('.py'):
        result = result[:-3]
    result = re.sub('[^a-z0-9.-]', '-', result)
    return result


@cache()
def _load_package_names():
    """Load PyDist's name → Debian distribution name pairs

    i.e. binary name without "python3-" prefix if it's set

    f.e.:
      SQLAlchemy → sqlalchemy
      sphinxcontrib_openapi → sphinxcontrib.openapi

    NOTE: this guesses name from dependency line which is not correct in all cases
    """
    result = {}
    try:
        data = load('cpython3')
    except Exception as err:
        log.warn('cannot load pydist names: %s', err)
        data = {}
    else:
        for key, details in data.items():
            a = details[0].dependency.replace('python3-', '')
            if ' ' in a:  # ignore version and other options (alternative deps?)
                a = a.split()[0]
            result[key.lower()] = a
    return result



# Copyright © 2015-2018 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging
from os import environ
from os.path import exists, join
from xmlrpc.client import ServerProxy
# from aioxmlrpc.client import ServerProxy  # TODO: package it in Debian
import aiohttp

from pypi2deb.decorators import cache
from pypi2deb.tools import pkg_name, execute

PYPI_JSON_URL = environ.get('PYPI_JSON_URL', 'https://pypi.org/pypi')
PYPI_XMLRPC_URL = environ.get('PYPI_XMLRPC_URL', 'https://pypi.org/pypi')
log = logging.getLogger('pypi2deb')


async def get_pypi_info(name, version=None):
    url = PYPI_JSON_URL + '/' + name
    if version:
        url += '/' + version
    url += '/json'

    async with aiohttp.ClientSession(trust_env=True) as session:
        try:
            response = await session.get(url)
        except Exception as err:
            log.error('invalid project name: {} ({})'.format(name, err))
            return
        try:
            result = await response.json()
        except Exception as err:
            log.warn('cannot download %s %s details from PyPI: %r', name, version, err)
            return
        return result


def parse_pypi_info(data):
    if not data:
        return {}

    for key, value in list(data['info'].items()):
        if value == 'UNKNOWN':
            data['info'][key] = ''

    info = data['info']
    result = {
        'name': info['name'],
        'version': info['version'],
        'description': info['description'].replace('\r\n', '\n'),
        'license_name': info['license'],
        'author': '{author} <{author_email}>'.format(**info),
        'homepage': info['home_page'],
    }
    if 'requires' in info:  # see f.e. qutebrowser
        result['requires'] = info['requires']

    summary = info.get('summary', 'FIXME').replace('  ', ' ')

    unwanted_prefixes = {'a', 'an', 'the', 'is', result['name'].lower()}
    while True:
        if ' ' not in summary:
            break
        prefix, rest = summary.split(maxsplit=1)
        if prefix.lower() in unwanted_prefixes:
            summary = rest
        else:
            break
    if summary.endswith('.'):
        summary = summary[:-1]
    result['summary'] = summary

    classifiers = data['info']['classifiers']
    result['interpreters'] = set()
    for i in classifiers:
        if i.startswith('Programming Language :: Python :: 3'):
            result['interpreters'].add('python3')
    # if 'Programming Language :: Python :: Implementation :: PyPy' in classifiers:
    #     result['interpreters'].add('pypy')
    if not result['interpreters']:
        # assume it's Python 3 only
        result['interpreters'].add('python3')

    return result


def parse_pkg_info(fpath):
    """Parse PKG-INFO file"""
    raise NotImplementedError()  # FIXME


async def download(name, version=None, destdir='.'):
    details = await get_pypi_info(name, version)
    if not details:
        raise Exception('cannot get PyPI project details for {}'.format(name))

    download_version = version or details['info']['version']

    package_urls = details['urls']
    if not package_urls:
        log.debug('missing release of %s %s on PyPI', name, download_version)
        raise Exception('missing release')

    try:
        release = next((i for i in package_urls if i['python_version'] == 'source'))
    except StopIteration:
        release = None

    if not release:
        available_files = ", ".join("{} (package type: {})".format(f["filename"], f["packagetype"]) for f in package_urls)
        other_versions_text = ""
        if not version and "releases" in details:
            other_available_versions = [v for v in details["releases"] if v != download_version]
            if other_available_versions:
                other_versions_text = " Some other versions found for {}: {}{}.".format(
                    name,
                ", ".join(other_available_versions[:5]),
                " and others" if len(other_available_versions) > 5 else ""
                )
            else:
                other_versions_text = " No other versions for {} found.".format(name)
        raise Exception('Source package for {} version {} not available on PyPI. Available files: {}.{}'.format(name, download_version, available_files, other_versions_text))

    orig_ext = ext = release['filename'].replace('{}-{}.'.format(name, download_version), '')
    if ext not in {'tar.gz', 'tar.bz2', 'tar.xz'}:
        ext = 'tar.xz'

    fname = '{}_{}.orig.{}'.format(pkg_name(name), download_version, ext)

    fpath = join(destdir, fname)
    if exists(fpath):
        return fname

    async with aiohttp.ClientSession(trust_env=True) as session:
        log.debug(f"fetching upstream tarball from {release['url']}")
        response = await session.get(release['url'])
        with open(fpath if ext == orig_ext else join(destdir, release['filename']), 'wb') as fp:
            data = await response.read()
            fp.write(data)

    if orig_ext != ext:
        cmd = ['mk-origtargz', '--rename', '--compression', 'xz',
               '--package', pkg_name(details['info']['name']), '--version', download_version,
               '--directory', destdir,
               '--repack', join(destdir, release['filename'])]
        # TODO: add --copyright-file if overriden copyright file is available
        await execute(cmd)

    return fname


@cache()
def list_packages(classifiers=None):
    client = ServerProxy(PYPI_XMLRPC_URL)
    if not classifiers:
        # packages = await client.list_packages()
        packages = client.list_packages()
        packages = dict.fromkeys(packages)
    else:
        # packages = await client.browse(classifiers)
        packages = client.browse(classifiers)
        # browse returns all versions, use the latest one only
        # TODO: use LooseVersion
        packages = dict(sorted(packages))
    return packages



# Copyright © 2021 Sandro Tosi <morph@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# PyGithub doesnt support (yet) asyncio, and there's only a prototype of a github
# client supporting it, so we're gonna do it the old way

import logging
from os.path import join, exists

import aiohttp
from github import Github
from github.GithubException import UnknownObjectException

log = logging.getLogger('pypi2deb')


async def github_download(name, github_url, version=None, destdir='.'):
    g = Github()
    repo_name = github_url.replace('https://github.com/', '').rstrip('/')
    log.debug(f"Calling github get_repo with arg {repo_name}")
    repo = g.get_repo(repo_name)

    if not name:
        name = repo.name

    try:
        tag_name = repo.get_latest_release().tag_name
    except UnknownObjectException:
        # Some projects do not use Github Releases, check the latest tag instead
        tag_name = repo.get_tags()[0].name

    if not version:
        # TODO: are there other special cases? vx.y.z tag gets rewritten as x.y.z
        version = tag_name.lstrip('v')

    # cant use this for now, chk https://github.com/PyGithub/PyGithub/issues/1871
    # download_url = latest_release.tarball_url
    # so let's "forge" the right URL here
    download_url = f"{github_url}/archive/refs/tags/{tag_name}.tar.gz"

    fname = f'{name}_{version}.orig.tar.gz'

    fpath = join(destdir, fname)
    if exists(fpath):
        return fname

    async with aiohttp.ClientSession(trust_env=True) as session:
        log.debug(f"fetching upstream tarball from {download_url}")
        response = await session.get(download_url)
        with open(fpath, 'ba') as fp:
            data = await response.read()
            fp.write(data)

    return fname



# Copyright © 2015-2018 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging

__all__ = ['load', 'dump']

try:
    import msgpack as _serializer
    NAMESPACE = 'P2D:'
except ImportError:
    NAMESPACE = 'P2D-j:'
    try:
        import simplejson as _serializer
    except ImportError:
        import json as _serializer
try:
    import redis
except ImportError:
    conn = None
else:
    conn = redis.Redis()
    try:
        conn.ping()
    except redis.ConnectionError:
        conn = None
if not conn:

    class _FallbackCache(dict):
        def setex(self, key, data, ttl=None):
            self[key] = data

        def get(self, key, default=None):
            return super(_FallbackCache, self).get(key, default)

    conn = _FallbackCache()


log = logging.getLogger('pypi2deb')


def load(key, default=None):
    result = conn.get(NAMESPACE + key)
    if result is None:
        return default
    try:
        return _serializer.loads(result, encoding='utf-8')
    except Exception as err:
        exc_info = log.level <= logging.DEBUG
        log.warn('cannot load cache (%s): %s', key, err, exc_info=exc_info)
        return default


def dump(key, data, ttl=3600):
    exc_info = log.level <= logging.DEBUG
    try:
        data = _serializer.dumps(data)
    except Exception as err:
        log.warn('cannot serialize cache (%s): %s', key, err, exc_info=exc_info)
        return
    try:
        conn.setex(NAMESPACE + key, data, ttl)
    except Exception as err:
        log.warn('cannot dump cache (%s): %s', key, err, exc_info=exc_info)



# Copyright © 2015-2026 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import sys
from os import environ
from os.path import join, dirname, abspath

TEMPLATES_PATH = environ.get('PYPI2DEB_TEMPLATES_PATH',
                             abspath(join(dirname(__file__), '..', 'templates')))
OVERRIDES_PATH = environ.get('PYPI2DEB_OVERRIDES_PATH',
                             abspath(join(dirname(__file__), '..', 'overrides')))
PROFILES_PATH = environ.get('PYPI2DEB_PROFILES_PATH',
                            abspath(join(dirname(__file__), '..', 'profiles')))
VERSION = '5.20260225'
# Add path to dh-python's private library
# (yeah, it's not stable enough to make it public one, fortunatly
# author of pypi2deb and dh-python know each other ;)
sys.path.append('/usr/share/dh-python/')
